/*
** Graph.c
** 
** Made by (Johnny Q. Hacker)
** Login   <solarion@johnathan>
** 


Copyright (C) 2008 Joseph Pingenot

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


**Many thanks to Davyd Madeley for the excellent Cairo  Tutorial
**  at http://gnomejournal.org/article/34/writing-a--using-cairo-and-gtk28

** Started on  Thu Jul 17 10:51:32 2008 Johnny Q. Hacker
** Last update Sun May 12 01:17:25 2002 Speed Blue
*/

#include <graph.h>
#include <stdlib.h>
#include <stdio.h>

G_DEFINE_TYPE(Graph, graph, G_TYPE_OBJECT);

static void graph_class_init(GraphClass *klass) {
  GObjectClass *gobject_klass;
  gobject_klass = G_OBJECT_CLASS(klass);
}

static void graph_init(Graph *graph) {
  graph->symbol = CIRCLE;
  graph->points = g_ptr_array_new();
  graph->lines = NULL;
}

Graph *graph_new(void) {
  return g_object_new(GRAPH_TYPE, NULL);
}

/*1 means you're stupid and passed NULL*/
int graph_add_graph_point(Graph* graph, struct graph_point *pt) {
  if(pt == NULL) return 1;
  g_ptr_array_add(graph->points, pt);
  if((graph->points->len == 1) && ((graph->lines == NULL) || (graph->lines->len == 0))) {
    graph->maxx = graph->minx = pt->x;
    graph->maxy = graph->miny = pt->y;
  }else{
    if(pt->x > graph->maxx) graph->maxx = pt->x;
    if(pt->x < graph->minx) graph->minx = pt->x;
    if(pt->y > graph->maxy) graph->maxy = pt->y;
    if(pt->y < graph->miny) graph->miny = pt->y;
  }
  return 0;
}

/*
 *
 *1 means failed to allocate point
 *NOTE: label is duplicated if not NULL (using g_strdup)
 */
int graph_add_point(Graph* graph, double x, double y, const GString *label) {
  struct graph_point *pt = (struct graph_point*)malloc(sizeof(struct graph_point));
  if(pt == NULL) return 1;
  pt->x = x;
  pt->y = y;
  pt->label = (label)?g_string_new(label->str):NULL;
  graph_add_graph_point(graph, pt);
  return 0;
}

int graph_add_graph_line(Graph* graph, struct graph_line *l) {
  if(graph->lines == NULL) {
    if((graph->lines = g_ptr_array_new()) == NULL) return 1;
  }
  g_ptr_array_add(graph->lines, l);
  if((graph->points->len == 0) && 
     (graph->lines->len == 1)) {
    graph->maxx = graph->minx = l->p0_x;
    graph->maxy = graph->miny = l->p0_y;
  }else{
    if(l->p0_x > graph->maxx) graph->maxx = l->p0_x;
    if(l->p0_x < graph->minx) graph->minx = l->p0_x;
    if(l->p0_y > graph->maxy) graph->maxy = l->p0_y;
    if(l->p0_y < graph->miny) graph->miny = l->p0_y;
  }
  if(l->p3_x > graph->maxx) graph->maxx = l->p3_x;
  if(l->p3_x < graph->minx) graph->minx = l->p3_x;
  if(l->p3_y > graph->maxy) graph->maxy = l->p3_y;
  if(l->p3_y < graph->miny) graph->miny = l->p3_y;
  return 0;
}


int graph_add_line(Graph* graph, double p0_x, double p0_y, double p1_x, double p1_y, double p2_x, double p2_y, double p3_x, double p3_y) {
  #ifdef DEBUG
  fprintf(stderr, "graph_add_line(%d, p0_x=%g, p0_y=%g, p1_x=%g, p1_y=%g, p2_x=%g, p2_y=%g, p3_x=%g, p3_y=%g\n", (unsigned int) graph, p0_x, p0_y, p1_x, p1_y, p2_x, p2_y, p3_x, p3_y);
  #endif
  struct graph_line *l;
  if((l=(struct graph_line*)malloc(sizeof(struct graph_line))) == NULL) return 2;
  l->p0_x=p0_x;
  l->p0_y=p0_y;
  l->p1_x=p1_x;
  l->p1_y=p1_y;
  l->p2_x=p2_x;
  l->p2_y=p2_y;
  l->p3_x=p3_x;
  l->p3_y=p3_y;
  return graph_add_graph_line(graph, l);
}
    
struct cxt {
  Graph *graph;
  struct graph_point *p0;
};
    
static void linear_interpolate(gpointer data, gpointer user_data) {
  struct cxt *cxt = user_data;
  struct graph_point *p3 = data;
  if(cxt->p0 != NULL) {
    double p1_x = (p3->x + 2*cxt->p0->x)/3.0;
    double p1_y = (p3->y + 2*cxt->p0->y)/3.0;
    double p2_x = (2*p3->x + cxt->p0->x)/3.0;
    double p2_y = (2*p3->y + cxt->p0->y)/3.0;
    graph_add_line(cxt->graph, cxt->p0->x, cxt->p0->y, p1_x, p1_y, p2_x, p2_y, p3->x, p3->y);
  }
  cxt->p0 = p3;
}
    
void graph_add_linear_connectors(Graph* graph) {
  struct cxt cxt;
  cxt.graph = graph;
  cxt.p0 = NULL;
  g_ptr_array_foreach(graph->points, &linear_interpolate, &cxt);
}

void graph_bezier_linear_to_cubic(double x0, double x3, double *x1, double *x2) {
  *x1 = (x3 + 2.0*x0)/3.0;
  *x2 = (2.0*x3 + x0)/3.0;
}

void graph_bezier_linear_to_quadratic(double x0, double x3, double *x1) {
  *x1 = (x3 + x0)/2.0;
}

void graph_bezier_quadratic_to_cubic(double x0, double x3, double *x1, double *x2) {
  *x2 = (x3-2.0*(*x1))/3.0;
  *x1 = (x0-2.0*(*x1))/3.0;
}
